# chrono - content management system
# Copyright (C) 2016-2023  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import json
import tarfile

from django.contrib.auth.models import Group
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from rest_framework import permissions
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response

from chrono.agendas.models import Agenda
from chrono.apps.export_import.models import Application, ApplicationElement
from chrono.manager.utils import import_site


class Index(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def get(self, request, *args, **kwargs):
        response = [
            {
                'id': 'agendas',
                'text': _('Agendas'),
                'singular': _('Agenda'),
                'urls': {
                    'list': request.build_absolute_uri(reverse('api-export-import-agendas-list')),
                },
            },
        ]

        return Response({'data': response})


index = Index.as_view()


def get_agenda_bundle_entry(request, agenda, order):
    return {
        'id': str(agenda.slug),
        'text': agenda.label,
        'type': 'agendas',
        'urls': {
            'export': request.build_absolute_uri(
                reverse('api-export-import-agenda-export', kwargs={'slug': str(agenda.slug)})
            ),
            'dependencies': request.build_absolute_uri(
                reverse('api-export-import-agenda-dependencies', kwargs={'slug': str(agenda.slug)})
            ),
            'redirect': request.build_absolute_uri(
                reverse('api-export-import-agenda-redirect', kwargs={'slug': str(agenda.slug)})
            ),
        },
    }


class ListAgendas(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def get(self, request, *args, **kwargs):
        response = [
            get_agenda_bundle_entry(request, x, i) for i, x in enumerate(Agenda.objects.order_by('slug'))
        ]
        return Response({'data': response})


list_agendas = ListAgendas.as_view()


class ExportAgenda(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def get(self, request, slug, *args, **kwargs):
        serialisation = Agenda.objects.get(slug=slug).export_json()
        return Response({'data': serialisation})


export_agenda = ExportAgenda.as_view()


class AgendaDependencies(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def get(self, request, slug, *args, **kwargs):
        agenda = Agenda.objects.get(slug=slug)

        def dependency_dict(element):
            if isinstance(element, Group):
                return {
                    'id': element.role.slug if hasattr(element, 'role') else element.id,
                    'text': element.name,
                    'type': 'roles',
                    'urls': {},
                }
            elif isinstance(element, Agenda):
                return get_agenda_bundle_entry(request, element, 0)
            return element

        dependencies = [dependency_dict(x) for x in agenda.get_dependencies() if x]
        return Response({'err': 0, 'data': dependencies})


agenda_dependencies = AgendaDependencies.as_view()


def agenda_redirect(request, slug):
    agenda = get_object_or_404(Agenda, slug=slug)
    return redirect(reverse('chrono-manager-agenda-view', kwargs={'pk': agenda.pk}))


class BundleCheck(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def put(self, request, *args, **kwargs):
        return Response({'err': 0, 'data': {}})


bundle_check = BundleCheck.as_view()


class BundleImport(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)
    install = True

    def put(self, request, *args, **kwargs):
        tar_io = io.BytesIO(request.read())
        agendas = []
        with tarfile.open(fileobj=tar_io) as tar:
            manifest = json.loads(tar.extractfile('manifest.json').read().decode())
            self.application = Application.update_or_create_from_manifest(
                manifest,
                tar,
                editable=not self.install,
            )

            for element in manifest.get('elements'):
                if element.get('type') != 'agendas':
                    continue
                agendas.append(
                    json.loads(tar.extractfile(f'agendas/{element["slug"]}').read().decode()).get('data')
                )
        # init cache of application elements, from manifest
        self.application_elements = set()
        # import agendas
        self.do_something(agendas)
        # create application elements
        self.link_objects(agendas)
        # remove obsolete application elements
        self.unlink_obsolete_objects()
        return Response({'err': 0})

    def do_something(self, agendas):
        if agendas:
            import_site({'agendas': agendas})

    def link_objects(self, agendas):
        for agenda in agendas:
            try:
                existing_agenda = Agenda.objects.get(slug=agenda['slug'])
            except Agenda.DoesNotExist:
                pass
            else:
                element = ApplicationElement.update_or_create_for_object(self.application, existing_agenda)
                self.application_elements.add(element.content_object)

    def unlink_obsolete_objects(self):
        known_elements = ApplicationElement.objects.filter(application=self.application)
        for element in known_elements:
            if element.content_object not in self.application_elements:
                element.delete()


bundle_import = BundleImport.as_view()


class BundleDeclare(BundleImport):
    install = False

    def do_something(self, agendas):
        # no installation on declare
        pass


bundle_declare = BundleDeclare.as_view()


class BundleUnlink(GenericAPIView):
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def post(self, request, *args, **kwargs):
        if request.POST.get('application'):
            try:
                application = Application.objects.get(slug=request.POST['application'])
            except Application.DoesNotExist:
                pass
            else:
                application.delete()

        return Response({'err': 0})


bundle_unlink = BundleUnlink.as_view()
